// Device: This is the building block class for all of World Machine.
// a device takes zero or more inputs, performs some operation,
// and produces zero or more data packets as ouptut.
// The type of data, and operation, is completely open.

#pragma once
#ifndef __DEVICE_H
#define __DEVICE_H

#include <vector>
#include "common.h"
#include "WMPacket.h"
#include "Globals.h"
#include "Parameter.h"
#include "Ports.h"

const int SOCKET_SIZE_X = 8;
const int SOCKET_SIZE_Y = 8;
const int PSOCKET_SIZE_X = 4;
const int PSOCKET_SIZE_Y = 6;

const int DEV_NAME_LEN = 24;

const int DEV_INPUT = 0;
const int DEV_OUTPUT = 1;
const int DEV_PARAMIN = 2;
const int DEV_MASK = 3;
const int DEV_COORD = 4;

const int DEV_PREVIEW_DONE = 1;
const int DEV_FINAL_DONE = 2;
const int DEV_SLEEPING = 4;

// World events that a device may or may not depend on. Right now just a few of these...
// used by isUpdatedOn to control if the device is reset or not on this event
const int DEV_EVENT_LIGHTING = 1;
const int DEV_EVENT_WATER = 2;
const int DEV_EVENT_COLORTABLE = 4;
const int DEV_EVENT_RENDER_EXTENTS = 8;

struct WorldCommonData;
struct RenderExtent;
class DeviceWorldBuilder;
class DeviceWorld;
class Port;
class Parameter;
class ParameterSet;
class LayoutUIClient;

enum DEVICESPACE { WORLDSPACE, LOCALSPACE };

enum PSTATE { P_ALL = -1, P_FINAL = 0, P_PREVIEW = 1, P_TEMP = 2 };

enum ACTIVATE_RESULT { ACTIVATE_SUCCESS = 1, ACTIVATE_FAIL = 0, ACTIVATE_CURRENT = -1, ACTIVATE_DISABLED = -2 };
/* BuildContext provides the context necessary for a device to function.
 The information passed includes the preview state, render extents, memory conservation,
 as well as a reporting mechanism to pass progress reports back to the Builder.

BuildContext gets passed to a device's Activate() and PreActivate() functions
	
*/
class BuildContext {
public:
	BuildContext() { builder = NULL; allow_device_threading = false; tiled = false; conserve_memory = false; pstate = P_PREVIEW; cache_full_build = false; };

	operator PSTATE&() { return pstate; };

	void SetWorldSize(RenderExtent &re, WorldCommonData &wpd);
	SizeData &GetWorldSize() { return sd; };
	CoordF    GetPanLoc() { return pan; };
	RectF GetRenderQuad();
	Rect4F GetRenderQuad4F();

	// Progress Reporting operations
	DeviceWorldBuilder *builder; // The Builder that is orchestrating the build sequence, that we're working for
	virtual void ReportDeviceBegan(Device *dev);				 // Reports when a device is to begin building
	virtual bool ReportProgress(Device *dev, int amnt);			 // Progress on a 0..100 scale
	virtual bool ReportProgress(Device *dev, int amnt, int max); // Progress normalize to 0..max
	virtual void ReportDeviceBuilt(Device *dev);				 // Reports when a device is done

	PSTATE pstate;					// Which build status this activation falls into
	bool tiled;						// flagged if this is a tiled build
	bool conserve_memory;			// use memory conservation
	bool allow_device_threading;	// if multithreading-capable devices should use worker pool threads
	bool cache_full_build;			// if true, build process will cache final-mode things
	SizeData sd;					// Resolution and scaling data
	CoordF pan;						// panning loc
private:
};


// Parameter Reference Accessors
// These are defined indirectly so that the same parameter names can be used
// both inside of the device itself, and inside it's custom dialog box, if any.
// defined as references so as to able to be both set and get
#define ParmIRef(x)		(params.getOwner()->GetIRef(x))
#define ParmFRef(x)		(params.getOwner()->GetFRef(x))
#define ParmBRef(x)		(params.getOwner()->GetBRef(x))

class Device
{
public:
	Device(void);
	virtual ~Device(void);

	friend class GraphOps;
	friend class DeviceWorldBuilder;
	friend class DeviceWorld;
	friend class GenericDlg;
	friend class BuildEngine;
	friend class PresetManager;

// Each device has a world that it belongs to.
	void SetWorld(DeviceWorld *world) { worldptr = world; };
	DeviceWorld *GetWorld() { return worldptr; };

// Serialization
	virtual bool Load(std::istream &in);	// Framework code for devices to serialize to disk. 
	virtual bool Save(std::ostream &out);
	virtual bool LoadSession(std::istream &in);
	virtual bool SaveSession(std::ostream &out);

	Device *Clone();		// Creates an independant copy of this device
	void Destroy();			// Delete the device using its internally defined deletion operator

// Miscellaneous UI related functions
	virtual dword Color() { return 0x07f7f7f; }; 
	virtual int GetWidth();
	virtual int GetHeight();
	virtual char *GetDescription() { return "This part can never be instanced";};
	virtual void  SetName(const char *name);
	virtual char *GetName();
	virtual char *GetTypeName() { return "Unknown Device"; };
	char *GetStaticTag() { return lifeptrs.nametag; };
	virtual int GetDeviceFamily() = 0;
	void SetDeviceID( int new_id) { device_uid = new_id; };
	int GetDeviceID() { return device_uid; };

	virtual char *GetInputName(int slot);
	virtual char *GetOutputName(int slot);
	char *GetParamInputName(int slot);		// Non-overridable since it directly interfaces with parameterset

// Device Workview functions
	Coord GetLoc() { return loc; };			// Device's workview location
	void SetLoc(Coord &newloc) { loc = newloc; };
	int	FindPort(Coord loc, int *io, int showstate);		// sets io ptr to io: 0 = input, 1 = output, 2 = param 3 = mask
	Coord GetPortLoc(int slot, int io);
	Coord GetPortLoc(Port *target_port, int io = -1);

// Layout View functions (optional!)
	virtual LayoutUIClient *GetLayoutUI() { return NULL; };

// Device origin / transformation functions
	CoordF getOrigin();						// Return origin modified by the coordinate input, if any
	CoordF getOriginRaw() { return origin; };// return unmodified device origin
	void   setOrigin(CoordF &o) { origin = o; };
	int	   getOriginType() { return origin_type; };
	void   setOriginType(int t) { origin_type = t; };

// Graph&Port operations
	void LinkClear();					// Clear all ports of any links
	int  GetLinkedPort(Port *linked_port, int io = -1);	// Return the slot of a port linked to this device
	Port *GetPort(int index, int io);	// Return a pointer to the given port
	bool FindPort(const Port *port, int *index, int *io);// Look for the specified port on this device and return its locale
	int  GetNextPort(int io);			// Returns the next unconnected port index of a type, or -1 for none
	bool CheckPort(int slot, int io);	// Is the specified port connected?
	int  GetPortStyle(int slot, int io);	// Return if the port is flagged special or not
	Port *in(int slot = 0) { if (slot < (int) inputs.size()) return inputs[slot]; else return NULL;};
	OutputPort *out(int slot = 0) { if (slot < (int) outputs.size()) return outputs[slot]; else return NULL; };
	Port *parmin(int slot = 0) { if (slot < (int) params_input.size()) return params_input[slot]; else return NULL; };
	Port *mask() { return mask_input; };	// Return the mask port
	Port *coord_port() { return coord_input; };	// Return the coordinate port

	int  highest_input();	// index of the highest connected port. -1 == no connections
	int  highest_output();	
	int  active_inputs();	// number of connected inputs/outputs
	int  active_outputs();
	int  active_param_inputs();	
	int	 max_inputs() { return (int) inputs.size(); };
	int  max_outputs() { return (int) outputs.size(); };
	int  max_parm()	{ return (int) params_input.size(); };

	// Notification when a wire is attached to a port for this device
	// default does nothing; sophisticated devices might accept several different input types on a given port and morph themselves depending on input type
	virtual void NotifyPortConnected(int slot, int io, bool is_connecting);	

// Device functionality -- these are the actual do-something methods
	virtual bool RunUI();				// Run the user interface for this part. Over-ride to show a custom dlg
	bool		 RunNameDialog();		// Run the user interface to set a new name for this part
	virtual void RunHelpDialog();		// Run the user interface to display a dialog with info about the device

	virtual ACTIVATE_RESULT ActivateSequence(BuildContext &context);	// Wraps call to PreAct, ActOnNeed, and PostAct
	virtual ACTIVATE_RESULT ActivateOnNeed(BuildContext &context);		// Calls Activate only if changed. Generally don't override this

	virtual bool PreActivate(BuildContext &context);		// Is called before Activate to setup inputs and other things
	virtual bool Activate(BuildContext &context) = 0;		// THE ALL IMPORTANT Activate call. Actually PERFORM the device action
	virtual void PostActivate(BuildContext &context);		// Called after Activate call. Generally don't override, but for some special purposes you may want to call the base class then perform your own cleanup

	virtual bool RunAuxAction(int index);		// Run extra actions defined by this sprocket. Returns TRUE if framework should re-calculate after action
	virtual int  getAuxActionSize();			// The number of actions defined
	virtual char *getAuxActionName(int index);	// The name of the extra action

	DEV_DATA &GetLifeVars() { return lifeptrs; };

	virtual int	Randomize(int newseed = 0);		// Randomize this device, if applicable.

// Parameter Controls
	int		GetI( const char *name);
	float	GetF( const char *name);
	bool	GetB(const char *name);

	int		&GetIRef( const char *name);
	float	&GetFRef( const char *name);
	bool	&GetBRef(const char *name);

	dword   &GetCRef(const char *name); // RGB Color

	ParameterSet &GetParams() { return params; };
	void SetParams(ParameterSet &other) { params = other; }; //params.setOwner(this); };
	WMPacket *RetrieveParamPacket( int slot );

	virtual void Preview();							// Make sure this part (and its ancestors) have a preview built
	virtual void Reset(PSTATE which_build = P_ALL); // Reset part to unprocessed state, destroying any packets residing inside
	virtual void Falsify();							// Flag this and all proceeding devices to need re-building

	WMPacket *GetFinishedPacket(PSTATE which_build, int slot = 0);	// retrieve a packet from the output buffer

	void		setEnable(bool _en) { enable = _en; };	// enable/disable this device
	void		setBypass(bool _by) { bypass = _by; };

	// Interrogatives
	bool		 isEnabled() { return enable; };	// Is this device toggled to disable mode?
	bool		 isBypassed() { return bypass; };	// Is this device bypassed?
	virtual bool isBuilt(PSTATE which_build);		// Is this device current under the current preview state?
	bool		 isAsleep() { return (processed & DEV_SLEEPING) != 0; };	// Is this device asleep (signal'd to not build because of child caching
	bool		 isGhosted() { return !worldptr; };	// device is ghosted if it doesn't have a current owner world

	// Interrogatives that are intrinsic to the class of the device
	virtual bool isUpdatedOn(int worldevent) { return false; }; // Device returns true if it needs to be recreated when the given world parameter changes
	virtual bool isRegistered() { return false; };	// If this device requires the registered version or not.
	virtual bool isViewable() { return true; };		// This device has a sensible output for viewing? 
	virtual bool isPermanent() { return false; };	// Should this device be unenableable & undeletable in the workview?
	virtual bool isPresetCapable() { return true; };// Allow presets for this device in the generic dialog manager?
	virtual bool isExplorable() { return true; };   // If this device is compatible with explorer mode or not
	virtual bool isMaskable() { return true; };		// Should this device have a mask input?
	virtual bool hasOrigin() { return false; };		// Does this device use the origin value?
	virtual bool hasLocalOverride() { return false; }; // Does this device allow local space?
	virtual bool isCheckPoint() { return false; };	// Is this device a building checkpoint

	// Hinting
	virtual bool getSpeedClass() { return 3; };		// Provide a rough estimate of processing speed of this device, relative to perlin noise (3). Higher is slower
	virtual int  estimateMemoryUsage();				// Provide the framework with an estimate of the memory required by this device given current settings, in KB

	void setMaskOverride(int value = 0);	// Temporarily override how this device outputs data
	int  getMaskOverride() { return mask_override; };

	virtual WMPacket *getDefaultOutput(PSTATE preview);		// Retrieve the output the device wants to show as a preview.
															// if Viewable is false, it will descend the graph until true
// ** Coordinate System Transformation functions **
//  take a coordinate and transform it according to a given coordinate system
	CoordF TransformCoord_HFToWorld(BuildContext &bi, CoordF &device_loc);
	CoordF TransformCoord_WorldToHF(BuildContext &bi, CoordF &world_loc);

protected:
	virtual void doClone(Device *other);			// Copy data from other to this

	virtual void MaskOutput(WMPacket *map, int slotnum, PSTATE preview);	// Overridable depending on when/what you want to mask against

	void doMaskOperation(WMPacket *_result, WMPacket *_original, WMPacket *_mask);	// Actual blending operation

// Data input/output/activation functions
	WMPacket *RetrieveMaskOriginal(PSTATE pstate);			// Retrieve saved original state of input 0 for masking
	WMPacket *RetrieveMask(PSTATE preview);					// Retrieve mask input data
	WMPacket *RetrieveDataConst(int slot, PSTATE preview);	// Retrieves the ACTUAL input data, not a copy! DO NOT DESTROY GIVEN DATA

	WMPacket *RetrieveData(int slot, BuildContext &context);// Retrieve main inputs data
	void StoreData(WMPacket *dat, int slot, BuildContext &context);	// Stores a completed HF in our output buffer

	bool Bypass(BuildContext &context);						// Perform a bypass of this device, if possible
	virtual void FalsifyImp();								// Flag proceeding devices to need re-building


	DEV_DATA	lifeptrs;			// make&destroy pointers for this sprocket
	DeviceWorld *worldptr;			// pointer to this sprocket's container
	Coord loc;						// Coordinate of this sprocket in world space

	bool enable;					// If this device is disabled from activation or not.
	bool bypass;					// If this device should be bypassed
	int  mask_override;				// 0 = normal masking
									// 1 = override mask on
									// 2 = override mask off

	void SetLinks(int in, int out, int packet_type = WMP_hfield);	 // Setup this sprocket's link vector
	void SetLinksFlags(int in, int out);// Setup port flags
	void ChangeLinks(int in, int out, int packet_type = WMP_hfield);// Change the # of links without destroying existing ones. Used by devices
	void ChangeLinkIn(int in, int packet_type = WMP_hfield);		// for on-the-fly modifications of the # of links
	void ChangeLinkOut(int out, int packet_type = WMP_hfield);

	virtual void AddPort(int flag, int io, int packet_type = WMP_hfield, int index = -1);
	virtual void RemovePort(int io, int index );

	void AddParam(Parameter &p, int flags = 0);	// Adds a parameter to paramset, and also creates an override port for it
	void RemoveParam(int index);				// Remove a parameter, delete its override port
	void ResetParams();							// Deletes all params and param ports
	ParameterSet			params;				// Set of user controllable, managed-parameters
 
	std::vector<Port*>		inputs;	// Ports that contain the Digraph links
	std::vector<OutputPort*>outputs;
	std::vector<Port*>		params_input;	// Parameter-override hookups

	Port *mask_input;				// Built-in mask input
	Port *coord_input;				// Built-in coordinate input

	int indegree;					// used by the topo sort algo only.
	char name[DEV_NAME_LEN];		// name of this device

	void setBuilt(PSTATE which_build);
private:
	int processed;

	int	origin_type;				// How this device handles space conversion.
									// 0 = worldspace - global origin only (default)
									// 1 = localspace - ignore global settings
	CoordF origin;					// the device's origin.

//	virtual WMPacket *RetrieveSlot(Port *request, PSTATE pstate, bool forcenoclone = false);	// Request packet for the calling sprocket
	WMPacket *original_input0[3];

	UINT device_uid;			// Device unique identifier. NOT globally unique - guarenteed unique only within the current world file.
};

#endif
