#pragma once

#include <deque>
#include <vector>
#include "common.h"

/*

A Peer-based model to help Devices perform within-device threading. 

Override the DeviceThreadHelper class by defining all of the appropriate virtual functions, 
and extending the DeviceThreadWork and DeviceThreadData classes.

You can design your extended class in any way you wish, but generally you will pass extended data and parameters from 
the device into the Helper instance.

Then from the outside:  
 Call PerformWork() to begin building. PerformWork will launch as many worker threads as your worklist and
 the threadpool will support, and then begin working itself on the worklist. Function returns when all tasks are completed and all
 worker threads are released back to the pool.


Progress Reporting:

 DeviceThreadHelper will also call context.ReportProgress() based upon the progress through the worklist, freeing your work function
 from directly concerning itself with this. The framework also WILL SOON handle parent cancel messages correctly for you!
*/

class BuildContext;
class Device;

// A Work packet. What this is depends on the device in question, but it should contain data pointing to a quanta of work
struct DeviceThreadWork {
	int order;
};

// Thread Data. Each active thread will instantiate a class derived from this, and pass to the Action and Merge funcs
struct DeviceThreadData {
};

class DeviceThreadHelper
{
public:
	DeviceThreadHelper(Device *_owner, BuildContext &_context, bool ordered);
	~DeviceThreadHelper(void);

	// Begins threaded device build. This is the only outside-accessible function!
	bool PerformWork();	//Returns only when built results are finished and merged, or user abort
						// (return value of true indicates successful build)
protected:
	BuildContext &context;
	Device *owner;

	bool noAutoReporting;	// Set to TUE in your constructor if helper should NOT automatically call context.ReportProgress()

// ** Work related functions. If you're using the provided work que, fill the queue using doCreateWork().
	virtual void doCreateWork() { 		progress_max = waiting_work.size(); }; // Implement these two function sif noWorkQue == false
	virtual void doDeleteWork(DeviceThreadWork *work) { };
	virtual bool workAvailable();	// Override this if noWorkQue == true!

// **** Action Operations -- Implement these for your device!
	virtual bool doActionFunc(DeviceThreadWork *work, DeviceThreadData *data) = 0;	// Perform build operation on a piece of work
	virtual bool doMergeFunc(DeviceThreadWork *work, DeviceThreadData *data) = 0;	// Merge results of work into final data

	// Create Local data on a per-thread basis. Override DeviceThreadData to define data and create/destroy it per-thread
	virtual DeviceThreadData *doThreadDataCreate() { return NULL; };
	virtual void			  doThreadDataDestroy(DeviceThreadData *data) { return; };
// **** End

// Functions to call 
	void addWorklist(DeviceThreadWork *work);	// User-callable: Call from doCreateWork to add items to the worklist
	bool is_reordering() { return order_returning_work; };
protected:
	virtual bool threadProgress(int progress);

	DeviceThreadWork *RetrieveWork(DeviceThreadData *data);
	bool StoreWork(DeviceThreadWork *work, DeviceThreadData *data);
	virtual int ThreadFunc(bool leader, bool allow_childspawn = false);


	int progress_max;
	volatile bool cancel;			// set to true 
	volatile int progress_cur;
	volatile LONG num_threads;		// # of set threads
	volatile LONG thread_counter;	// # of active threads as measured by them
	static UINT staticThreadFunc(void *classptr);
	std::deque<DeviceThreadWork*> waiting_work;

	bool order_returning_work;
	std::vector<DeviceThreadWork*> returning_work;
	int returning_waitfor;
	void RetireWaitingWork(DeviceThreadData *data);

	CRITICAL_SECTION critsec;
};

// ThreadHelperScanline is a specialization for devices that do scanline-based processing, and can thus split up work on a single heightfield
// among threads and work on the same data without clobbering each other. All that is necessary to implement from this base class is the
// doActionFunc method, which will get a piece of work with a set range of scanlines, and has simply to call the building class to make them.
struct ThreadScanlineWork : public DeviceThreadWork {
	Coord range;
};

class ThreadHelperScanline : public DeviceThreadHelper {
public:
	ThreadHelperScanline(Device *_owner, BuildContext &_context, int _sections = 16);

	void setScanlineRange(int min, int max) { total_range.x = min; total_range.y = max; };

	virtual void doCreateWork();
	virtual void doDeleteWork(DeviceThreadWork *work) { ThreadScanlineWork *trw = (ThreadScanlineWork *) work; delete trw; };
	// Merge occurs automatically since we're working on same packet but different scanlines
	virtual bool doMergeFunc(DeviceThreadWork *work, DeviceThreadData *data) { return true; };

	// Implement this!
	virtual bool doActionFunc(DeviceThreadWork *work, DeviceThreadData *data) = 0;	// Perform build operation on a piece of work

protected:
	Coord total_range;
	int sections;
};
