#pragma once
#if !defined(AFX_SPROCKETWORLD_H__2A35E6AB_FC4C_40D6_B793_ADAE5EBD19EC__INCLUDED_)
#define AFX_SPROCKETWORLD_H__2A35E6AB_FC4C_40D6_B793_ADAE5EBD19EC__INCLUDED_

#include "common.h"
#include <vector>
#include <queue>

#include "device.h"
#include "Globals.h"

enum WM_DIM_TYPE { UNIT_WM, UNIT_METER };

const float WM_to_KM_ratio = 8.0f; // 8km in 1wu

/////// World Size and creation specs /////////

// Contains all information on global state of the world; things that are NOT dependant on render extents
struct  WorldCommonData {
	float vert_scale;
	bool allow_nonsquare;	// Allow non-square world extents
	bool force_even_ratio;	// Force non-square extents to be even-ratio'd
	WM_DIM_TYPE dimension_type;	// 0 = WM units
								// 1 = meters
	bool use_altitude_dimension;// Use units vertically or dimensionless
};

const int TILED_FLAG_ALLOW = 1;			// Enabled tiled building for this render extent
const int TILED_FLAG_USESUBSECTION = 2;	// Use subsectional information
const int TILED_FLAG_NAME_PAD = 4;		// Pad the naming strings to full width
const int TILED_FLAG_CUSTOMRES = 8;		// Using custom resolution (UI helper)
const int TILED_FLAG_CUSTOMNUM = 16;	// Using custom numtiles (UI helper)
const int TILED_FLAG_MERGEOUTPUT = 32;	// At Merge-up stage, combine all tiles into a single file
const int TILED_FLAG_EXTRA_EDGE_TILES = 64;	// Create additional tiles as needed to create perfectly-blended output tiles

// Contains information on the tiled building
struct TiledBuildData {
	TiledBuildData();

// assistance functions to query status flags
	bool testMergeTiles() const { return flags & TILED_FLAG_MERGEOUTPUT; };
	bool testUseShared() const { return !testMergeTiles() && share_edges; };

	Coord output_start;		// tile coordinate to start numbering with (usually 0,0 or 1,1)
	Coord numtiles;			// number of tiles per side
	int tileres;			// max res of each tile
	int blending_div_amount ;// amount extra to blend between tiles
	bool square_tiles;		// force square tiles, may slightly overlap render area
	bool share_edges;		// make tiles share vertices on edges
	bool flipYorder;		// tile in opposite Y ordering
	Rect4 subsection;		// partial output subsection data (optional)
	char tilestr[32];		// tile append naming string
	int flags;				// validity of several special things

	int getBlendingRegion() {
		if (blending_div_amount >= 5)
			return 0;
		int blend_region = tileres >> blending_div_amount;
		return blend_region;
	}
};

// Saved render extents structure
const int RE_STATUS_LOCKED = 1;

struct RenderExtent {
	RenderExtent(const char* _name);
	RenderExtent();
// functions
	bool isLocked() { return status & RE_STATUS_LOCKED; };
	void EnforceSquare();
// data
	Rect4F loc;					// worldspace location of this render extent
	int  resolution;			// resolution of max dimension
	char name[32];
	int status;
// tiled data
	TiledBuildData tbd;
};

////////////////////////////////////////////

struct  BrokenLink {
	bool isFrom;		// true if the "inside" device is the From one
	Device *from;
	Device *to;
	int from_port, to_port;
	int from_io, to_io;
	int	datatype;		// type of the "inside" port
};

struct GroupStruct {
	Rect4F	area;
	std::string name;
	std::string desc;
	DWORD color;
	int alpha;
	bool floating;		// group can be floated over other groups rather than pushes
};

// Manually routed wires, consisting of the starting and ending ports, and a list of the routing points
struct WireRoute {
	WireRoute() { a = NULL; b = NULL; state = 1; };
	Port *a, *b;
	std::vector<int> pointlist;	// list of route points, consisting of their index in the routepoint container
	int state;		// status of the Wire Route. 0 = invalid, ready to overwrite, 1 = in use
	bool usesPoint(int rp);
};

struct RoutePoint {
	RoutePoint() : loc(0,0) { state = 0; };
	Coord loc;
	int state;		// status of the RoutePoint. 0 = invalid, 1 = in use
};


class MacroDev;
class Output;

class DeviceContainer {
};

class DeviceWorld : public DeviceContainer
{
public:
	friend class GraphOps;
	friend class DeviceWorldBuilder;
	friend class BuildEngine;
	friend class MacroDev;

	DeviceWorld(void);
	virtual ~DeviceWorld(void);

	Device *LoadDevice(std::istream &in) const;
	virtual int  Load(std::istream &in);
	virtual bool Save(std::ostream &out);
	virtual bool Merge(std::istream &in, std::vector<Device*> *merged_devices = NULL);				// Merge contents of another world to tihs one

	DeviceWorld &operator=( const DeviceWorld & );
	void SetName(const char* _name);
	const char *GetName();

// Basic Device operations
	void Clear();
	int  NumDevice() { return int(parts.size()); };
	void AddDevice(Device *part, Coord &loc);
	int  AddDeviceAtIndex(Device *part, Coord &loc, int index);
	void DelDevice(Device *part);
	void DelDevice_StayLinked(Device *part);
	Device *GetDevice(int index) { if (index >= int(parts.size())) return NULL; return parts[index]; };
	int FindDeviceIndex(Device *dev);		// Find index # of a given device
	int GetDeviceIDNext() { return device_uid_next++; }; // Retrieves and increments the worldfile-specific device ID specifier. Automatically used by various parts of WM..

	Device *Find(Coord &loc);	// Find if a device is present at given coordinates
	Device *Find(int uid);		// Find a device, if present, with the given unique identifier

	std::vector<Device*> Find(Rect &box);
	std::vector<Device*> FindAllParts() { return parts; };
	std::vector<Device*> FindAllFamily(int type); // Return a vector of all devices of a particular family in this world

// Device Link ops
	bool MakeLink(Port *P1, Port *P2);							// Link together two given ports
	bool BreakLink(Port *P1, Port *P2 = NULL);					// Break the link(s) from a given port
	bool BreakLink(Device *a, int port, int io);				// Break link at specified dev/slot/iotype
	bool UnWireDevice(Device *part);	// Attempts to remove the device from the network, maintaining integrity (reconnect neighbor parts together)
	bool CheckCycle();		// Check for existance of cycle in graph

	bool AttachGraph(std::vector<Device*> &subparts);	// Add a block of (self-attached) parts to the graph
	std::vector<BrokenLink> DetachGraph(std::vector<Device*> &subparts, bool break_links = true);	// Break all links between subgraph and graph at large

	static void CloneSubGraph(const std::vector<Device*> &parts, std::vector<Device*> &dest);	// Duplicates a group of parts and returns their clones

// Routing ops
	int NumRoutes();
	WireRoute &GetWireRoute(int i);				// index into routelist
	WireRoute *FindWireRoute(Port *a, Port *b);	// find a route if it exists
	WireRoute &AddWireRoute(Port *a, Port *b, int coordinate, int location = -1);
	bool DeleteWireRoute(Port *a, Port *b);		// delete the route for a given wire if it exists

	int NumRoutePoints();
	int AddRoutePoint(Coord loc);
	Coord GetRoutePoint(int index);	// route point index
	bool isRoutePointValid(int index); // check if the route point is currently in use (valid)
	void ChangeRoutePoint(int index, Coord newloc);
	void DeleteRoutePoint(int index, bool remove_others = true);	// if remove_others is false, will not remove the route point if any others are using it..
	bool ChangeRouting(WireRoute &route, int routepoint, int new_routepoint);	// change a routepoint in a route from one to another. false if RP doesnt belong to the wire
	bool DeleteRouting(WireRoute &route, int start, int end, bool preserve_last);	// Deletes route points between two routepoints. false if RPs don't belong to this wire

// Device group ops
	void AddGroup(GroupStruct &newgroup) { groups.push_back(newgroup); };
	void DelGroup(int i) { groups.erase(groups.begin() + i); };
	GroupStruct &GetGroup(int i) { return groups[i]; };
	int numGroups() { return int(groups.size()); };

	int inGroup(Device *dev);	// Return index of the group the device belongs to, or -1 if none
	int inGroup(Coord &loc);	// Return index of the group the coordinate belongs to, or -1 if none

	void ResetMaps(PSTATE which = P_ALL);	// Remove 

	void setMemUseMode(int local_copy) { keeplocalcopy = local_copy; }; 
	int  getMemUseMode() { return keeplocalcopy; }; 

	int GetPacketMemory(int type);	// Return use, in KB, of a packet at the given world size
// World Updating
	void RandomizeWorld();
	bool UpdateWorldOnEvent(int deviceevent);	// Returns true if any device was reset/falsified

	unsigned long long GetWorldUID() { return world_state_id; };
	void IncrementWorldUID() { world_state_id++; };


// Render Extents and data

	SizeData GetWorldSize();

// Retrieve the current render extents
	RenderExtent &GetCurrentRenderExtent();
	int GetCurrentRenderExtentIndex();
// Retrieve tiling information for current render extent
	TiledBuildData &GetTileData() { return GetCurrentRenderExtent().tbd; };
	void SetTileData(TiledBuildData &_td) { GetCurrentRenderExtent().tbd = _td; };

// Management of saved extents
	int numRenderExtent() { return int(extents.size()); };
	RenderExtent &GetRenderExtent(int index);
	void ActivateRenderExtent(int index);	// Make an extent active
	void UpdatedRenderExtents();

	void AddRenderExtent(const RenderExtent &re);
	bool DelRenderExtent(int index);


	WorldCommonData &GetWorldParams()		{ return wdata; };
	void SetWorldParams(WorldCommonData &wd);

// Project default folder options (retrieve default folder, convert from a rel path to an abs)
	std::string &GetDefaultPath();
	std::string CreateAbsolutePath(std::string &rel_path);
	std::string CreateRelativePath(std::string &abs_path);
	bool path_uses_TMD_path;

// Water level
	float water_level;
	bool use_water;

// Sunlight environment (not saved as of right now *** )
	Vector3 sunlight_dir;

// Project information
	bool use_session_file;
	bool show_startinfo;
	std::string default_path;
	std::string info_author;
	std::string info_description;

// Scaling defaults
	int default_scaler;
	float default_scaler_parameter;


protected:
// New version: All data on current settings are stored in the deviceworld.
// resolution matching is strictly enforced between the different worlds.
	WorldCommonData wdata;

	// set of saved render extents
	std::vector<RenderExtent> extents;
	int using_extents;

    int preview_resolution;
	int keeplocalcopy;

	std::vector<Device*> parts;

	// comment box zones in the source world
	std::vector<GroupStruct> groups;

	// Routing wires and routing Points
	std::vector<WireRoute> wireroutes;
	std::vector<RoutePoint> routepoints;

private:
	char name[DEV_NAME_LEN];					// Name of this world
	unsigned long long world_state_id;			// a UID, incremented with a random number each time the world is changed
	int device_uid_next;						// counter for assigning device UIDs

	// protect copy constructors
	DeviceWorld(const DeviceWorld &other);		// copy construction

	WireRoute *getNextAvailableRoute();
	bool cleanRoutes(int point = -1);	// make sure this (or all) point is set to invalid, if unused..
};

#endif
