// The heightfield class wraps accesses to a basic hfield grid.
// It also includes some special things, such as a global HF store,
// "nametags", and a few other things.
//
//////////////////////////////////////////////////////////////////////

// Note that the sizes in HField are the "total" sizes, ie including the scaled for detail.
// to transform from total to real size, you must multiply by the detail scalar.
// GetRealSize() does this for you.

#if !defined(AFX_HFIELD_H__5095FCF1_CB94_4229_B855_0F89F8A35EE0__INCLUDED_)
#define AFX_HFIELD_H__5095FCF1_CB94_4229_B855_0F89F8A35EE0__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\core\common.h"
#include "..\core\WMPacket.h"
#include "..\core\globals.h"
#include <vector>


/** HField : The basic Heightfield class
 * -Note that heightfields should *NOT* be created on the stack (locally). You must acquire a HF
 *  through the global factory method this ensures that 
 *  the proper heap is being used for memory alloc across DLLs.
 * -
 **/
class HField : public WMPacket
{
public:
	HField();			// Protect constructor/destructor
	HField &operator=(const HField &other);

	virtual int getType() { return WMP_hfield; };
	virtual char *getTypeName() { return "Heightfield"; };
	virtual dword Color() { return 0x00000; };

	virtual WMPacket *Clone();			// Return a copy of this WMPacket-derived packet
	virtual bool Load(std::istream &input);	// Load from stream input
	virtual bool Save(std::ostream &output);	// Save a copy of this data to stream output

// initializes a Heightfield's data store to the size specified by the SizeData structure
// after initialization the heightfield is ready for read/write operations
	void Init(SizeData _size);
	void InitFrom(HField &other);		// Match other's attributes, but DOES NOT copy data contents! Use copy constructor if you want that
	virtual void Clear();

// Basic access methods. These are in order from fastest to slowest methods for accessing heightfield
// memory.
	float *GetDataPtr();			// Return a pointer to heightfield memory. USE WITH CAUTION!!!
	inline float &operator[](const int index);	// low overhead
	inline float value(const Coord loc);		// fastest coordinate-based accessor
	inline float &operator[](const Coord loc);
	inline float &operator[](const Vector2 loc);
	float valueSafe(const Coord loc);	// Clips coordinates so that out-of-range access works (slower)
	float valueSafeZero(const Coord loc);// same, but out-of-range locs return 0 height
	float valueF(const CoordF loc);		// Get bilinear-interpolated value. (slow)
	float valueSafeF(const CoordF loc);	// Interpolated, clipped retrieval. (slowest)

// Heightfield size accessor functions
	inline int w() const { return sd.size.x; };
	inline int h() const { return sd.size.y; };
	inline int area() const { return sd.size.x * sd.size.y; };
	inline Coord getsize() const { return sd.size; };
	inline CoordF getworldsize() const { return sd.scalesize; };

	virtual int getHeapUsage() { if (!data) return 0; return sd.size.x * sd.size.y * sizeof(float);	};

	int getIndex(const Coord &loc);	// Convert a coordinate into an index value
	int getStride();				// the stride of the heightfield (normally ==  w() )

// Basic HF Operations
	CoordF getRange();								// Returns range of heightfield data, x=min, y=max
	void Normalize();								// Rescale data to lie between 0..1
	void ClampRange();								// Clamp data to lie 0..1. 
	float getVScale() { return sd.vert_scale; };	// Vertical scale of this heightfield
	float getDScale() { return sd.detail_scale; };	// Detail scale (pixels/WMu)
	SizeData getSizeData() { return sd; };
	void setSizeData(SizeData _sd) { sd = _sd; };

// Some more advanced heightfield info operators
	float slope(const Coord loc);
	float slopeF(const CoordF loc);	// Get bilinear sampled value
	Vector2 slopeVector(const Coord &loc, bool zmultiply = true);

	friend class HFieldImporter;
	friend class HFieldExporter;

protected:
	HField(const HField &other);			// copy construction

	virtual ~HField();
	virtual void Kill();		// Call to delete its own type

	virtual void clearContents();
	virtual bool isTypePageable() { return true; };

	SizeData sd;
	Coord &size;			// Reference to SD size member, for quick use
	float *data;
};

// The inline data accessors
inline float HField::value(const Coord loc) {
	return data[((int)loc.y*size.x) + (int)loc.x];
};

inline float &HField::operator[](const int index) {
	return data[index];
}

inline float &HField::operator[](const Coord loc) {
	return data[(loc.y*size.x) + loc.x];
}

inline float &HField::operator[](const Vector2 loc) {
	return data[(ffi(loc.y)*size.x) + ffi(loc.x)];
}

inline float *HField::GetDataPtr() {
	return data;
}

// Safe Accessor inline function to provide a safe way to cast WMPacket to HField
// in debug mode it will assert if the packet is not a HF
inline HField *HF(WMPacket *pack) { 
	if (pack) {
		bool ok = pack->getType() == WMP_hfield;
		if (!ok)
			return NULL;
	}
	return (HField*) pack; 
};

inline HField *GetEmptyHF() {
	HField *hf = HF(WMGlobal->GetPacket(WMP_hfield));
	return hf; 
};

inline HField *GetNewHF(SizeData sd) {
	HField *hf = HF(WMGlobal->GetPacket(WMP_hfield));
	hf->Init(sd);
	return hf; 
};

void colorEncode(const float grey, byte &rg, byte &b);

#endif // !defined(AFX_HFIELD_H__5095FCF1_CB94_4229_B855_0F89F8A35EE0__INCLUDED_)
