#pragma once
/*

The Parameter class 


*/


#include "common.h"
#include <vector>

class Device;

const int PARAM_NAME_SIZE = 32;
const int GROUP_NAME_SIZE = 32;

// Widget types. This allows you some control over how your parameters are controlled generically
const int WIDGET_SLIDER = 0;					// provide a slider (default)
const int WIDGET_LOGSLIDER = 1;					// logarithmic scaling on the slider
const int WIDGET_EDITONLY = 2;					// provide only an edit box (no slider
const int WIDGET_SLIDER_FLAG_METERS = 512;		// display slider values in WM units or kilometers as per current preference
const int WIDGET_SLIDER_FLAG_HEIGHT = 256;		// Specify as a WM vertical height unit (ie WM 0..1 or m/km as per pref)

// The types of values that a parameter can be
enum V_TYPE { V_INT, V_FLOAT, V_BOOL, V_PTR, V_ENUM, V_STR, V_COLOR };

typedef void (*VPtrType)(Device *owner);

// The value of a parameter. Access the given type according to the V_TYPE of the parameter
union VAL {
	VAL() { };
	VAL(int _i) { i = _i; };
	VAL(float _f) { f = _f; };
	VAL(bool _b) { b = _b; };

	int i;
	float f;
	bool b;
	VPtrType vptr;		// in 64bit system, will make sizeof(VAL) == 64bits
	dword color;
};

class EnumType {
public:
	char name[PARAM_NAME_SIZE];
};


// for type V_PTR, the parameter is an action. The pointer stored is expected to be the
// address of a standard call function that accepts one void* parameter. The framework will pass
// in the Device* pointer of the current device to the supplied function. Your function can
// then call a member function of the device, or do whatever other action is desired.

const int PARM_FLAG_HIDDEN		= 1;	// Do not assign a port or show this param
const int PARM_FLAG_DEPRECATED	= 2;	// Do not assign a port or allow user to change this param
const int PARM_FLAG_CLIP		= 4;	// Instead of scaling the override value to min..max, clip it
const int PARM_FLAG_READONLY	= 8;	// Don't allow the user to modify this parameter. (Usually used for feedback to the user)
const int PARM_FLAG_CENTER		= 16;	// Center the parameter instead of left-justifying it
const int PARM_FLAG_NODELETE	= 32;	// Dont allow the user to delete this parameter
const int PARM_FLAG_NOPORT		= 64;	// Do not assign a port to this param
const int PARM_FLAG_DISABLED	= 128;	// Disabled by Dependency

const unsigned char PARAM_SYNCH = 0x0ce;

class Parameter {
public:
// generic constructors
	Parameter();
	Parameter(const char *_name);
// specific constructors
	Parameter(const char *_name, int _value, int _min, int _max, int _wtype = WIDGET_SLIDER);
	Parameter(const char *_name, float _value, float _min, float _max, int _wtype = WIDGET_SLIDER);
	Parameter(const char *_name, bool _value);
	Parameter(const char *_name, VPtrType _ptr);	// action (button) parameter
												// see notes above regarding call procedure
	Parameter(const Parameter&);
	~Parameter();
	Parameter& operator= (const Parameter& other);

	bool Load(std::istream &in);
	bool Save(std::ostream &out);

	bool LoadValue(std::istream &in);	// Assumes everything except the "value" parameter is already setup
	bool SaveValue(std::ostream &out);	// Write out only Value

	void CopyStripped(const Parameter &other);	// Copy only the vitals; dont copy the strings. Faster.

	void ClampValue();	// Enforces that the current value is between min..max

	int getInt()		{ assert(type == V_INT || type == V_ENUM); return value.i; };
	float getFloat()	{ assert(type == V_FLOAT); return value.f; };
	bool getBool()		{ assert(type == V_BOOL); return value.b; };
	dword getColor()	{ assert(type == V_COLOR); return value.color; };

	void setInt(int a) { value.i = a; };
	void setFloat(float f) { value.f = f; };
	void setBool(bool b) { value.b = b; };
	void setColor(dword c)	{ value.color = c; };

// enums can be integer-valued only. You must first setup the range before setting the enumations
	void SetupEnum(int entries = -1); // optional give # of entries here
	int  numEnum();
	bool setEnumString(const int val, const char *str);
	char *getEnumString(const int val);

	void setHelpString(const char *str);	// description of this parameter
	const char *getHelpString();

// Get the flag bits
	int  getFlags()			 { return flags; };
	void setFlag(int _flags) { flags = _flags; };
// Check some specific flag bits
	bool isVisible()			 { return !(flags & (PARM_FLAG_HIDDEN | PARM_FLAG_NOPORT | PARM_FLAG_DEPRECATED)); };
	bool isEnabled()			 { return !(flags & PARM_FLAG_DISABLED); };

	VAL value;					// the current value
	VAL min, max;				// the range of acceptable values
	V_TYPE type;				// the type of VTYPE
	int widget_type;			// how should it be shown & adjusted?
	char name[PARAM_NAME_SIZE];	// name of this parameter
private:
	int flags;
	std::vector<EnumType> names;// storage of enumations
	char *help_str;				// pointer to help text for this parameter, if any
};

class ParamGroup {
public:
	void setName(const char*_name);

	char name[GROUP_NAME_SIZE];	// group name
	int group_start;			// which param starts this group
	int group_end;				// which param ends this group
};

enum DEPENDS_TYPE { VALUE_EQUAL, VALUE_CHANGE };

struct DependsPair;

typedef void (*DEPENDS_FUNC)(Device *dev, DependsPair &depair);

struct DependsPair {
public:
	int dependsOwner;		// the index of the parameter that will be changed
	int dependsOn;			// the index of the parameter that this one depends ON
	VAL value;				// the value that should trigger this dependence. The type of VAL is the type of dependsOn
	DEPENDS_TYPE type;		// the type of dependence
	DEPENDS_FUNC func;		// optional function to call upon dependency update
};

class ParameterSet
{
public:
	ParameterSet(void);
	~ParameterSet(void);

	bool Load(std::istream &in);
	bool Save(std::ostream &out);

	void AddParam(Parameter &p);
	bool DelParam(int index);
	bool ReplaceParam(const Parameter &p);
	bool ReplaceParam(int index, const Parameter &p);	// replace specified parameter with new one
	void ResetParams();

	Parameter *GetParam(int index);
	Parameter *GetParam(const char *name);
	int		   GetParamIndex(const char *name);

	int		getInt(int i); 
	float	getFloat(int i);
	bool	getBool(int i);
	int		getInt(const char *name);
	float	getFloat(const char *name);
	bool	getBool(const char *name);

	int size() { return (int) params.size(); };
	void setOwner(Device *_spr) { owner = _spr; };
	Device *getOwner() { return owner; };

// Param grouping stuff
	void AddGroup(ParamGroup &group) { groups.push_back(group); };

// Parameter Dependences.. allows for a parameter to react to the state of another parameter
	bool ProcessDependencies(const int changedindex = -1);
	bool AddDependence(const char *dependsOwner, const char *dependsOn, DEPENDS_TYPE type, VAL dependsval, DEPENDS_FUNC func = NULL);

	bool allow_noncurrent_parms;	// ON parameteset load, load all parameters, or only those that currently exist
private:
	Device *owner;
	std::vector<Parameter> params;
	std::vector<DependsPair> dependence;
	std::vector<ParamGroup> groups;
};