/*/////////////////////////////////////////////////////////////////////

  WMPacket is the base class that encapsulates information transfer between
  machines. Most common subclasses are HField and VField. (Heightfield and Voxelfield)

  However any information that needs be passed between devices can be derived from
  WMPacket, even allowing for User-defined information transfer.

  For example, other than just transferring entire HFs, you could have as information
  a scalar, a vector, or a full color RGB image. This allows you to radically expand the
  capabilities of WMachine, in principle allowing all sorts of texture & image generation
  and manipulation.


  Note that the display class may not know what to do with them all. It will try to
  display the data in a manner appropriate, but if it doesn't know how to handle something
  and there are no custom display plugins registered for that packet type, it will remain
  undisplayed.


/////////////////////////////////////////////////////////////////////*/

#if !defined(AFX_WMPACKET_H__90A4604E_7B5D_4668_AA28_ECCC0ABCCA34__INCLUDED_)
#define AFX_WMPACKET_H__90A4604E_7B5D_4668_AA28_ECCC0ABCCA34__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "common.h"
#include "globals.h"

const int WMP_ANY		= 0;
const int WMP_parameter	= 1;			// Parameter Packet
const int WMP_scalar	= 2;
const int WMP_hfield	= 4;			// Heightfields
const int WMP_rgb		= 8;			// RGB Bitmaps
const int WMP_volfield	= 16;
const int WMP_coord		= 32;			// Coordinate Packet
const int WMP_text		= 64;			// Text
const int WMP_combo		= 128;
const int WMP_mesh		= 256;			// Mesh
const int WMP_reserved6	= 512;
const int WMP_reserved7 = 1024;
const int WMP_reserved8 = 2048;
const int WMP_reserved9 = 4096;
const int WMP_reserved10= 8192;
// If you want to create new, user-defined packet types, CONTACT ME FIRST.
// I will give you the bitcode to use to prevent collisions with any other packet type
const int WMP_PARAMS	= WMP_parameter | WMP_scalar;

// Packet Display Hints. Used for better matching display only
const int WMP_hint_landscape = 0; // treat as a landscape
const int WMP_hint_mask		 = 1; // treat as a mask
const int WMP_hint_nolight   = 2; // do not apply lighting; data may be prelit

std::string getPacketTypeString(const int type, const std::string concat_str = ", ");

class WMPacket  
{
public:
	virtual WMPacket &operator=(const WMPacket &other);

	virtual int getType() = 0;
	virtual char *getTypeName() = 0;
	virtual WMPacket *Clone() = 0;			// Return a copy of this WMPacket-derived packet

	virtual bool Load(std::istream &input) = 0;	// Load from stream input
	virtual bool Save(std::ostream &output) = 0;	// Save a copy of this data to stream output

	virtual void Clear();					// Clear contents of this packet

	virtual dword Color() = 0;				// Returns a 32bit color (doubles as an ID word) for this type
											// of data.

	int getDisplayHint() { return hint; };		// Hints to WM how to display this HF. 
	void setDisplayHint(int _hint) { hint = _hint; };

	int AddRef();						// increment reference counter
	int Release();						// Decrease the ref count by one, deleting if appropriate.
	int getRefCount() { return ref_count; };
	void SetNameTag(const char *name);	// Tag this packet with a name
	char *GetNameTag();					// Return packet's name

	virtual int getHeapUsage() = 0;		// Return the number of bytes of WM heap memory this packet is using

	bool PageOut(bool clearcontents = true);
	bool PageIn();
	bool isPaged();

	bool isPageable() { return page_able > 0; };
	bool setPageable(bool available);				// set underlying pageability of the packet

	bool requestAccessibility(bool available);		// called when a view or other part needs to temporarily ensure the packet is paged in

protected:
#define NAMELEN		32

	friend void WM_FreePacket(WMPacket *wmp);
	WMPacket();
	WMPacket(const WMPacket &other);		// copy construction
	virtual ~WMPacket();
	virtual void Kill() = 0;		// Call to delete its own type
	virtual void clearContents() = 0;	// Called to reset heightfield contents during page out operation
	virtual bool isTypePageable() { return false; };

	int hint;

	int  page_able;	// page_able is a semaphore that controls pageable status
	bool page_state;
	char page_swapname[NAMELEN];

	int ref_count;			// Reference counting for the packet
	char nametag[NAMELEN];	// nametag of this packet
};

#endif // !defined(AFX_WMPACKET_H__90A4604E_7B5D_4668_AA28_ECCC0ABCCA34__INCLUDED_)
