#include "stdafx.h"
#include "inverter.h"
#include "..\core\HField.h"

// You must furnish your own Maker and Killer functions. The Maker should (obviously) be unique for each device,
// while the Killer can be shared among plugins in the same DLL. The basic idea is to get around the sometimes-messy
// problem of DLL<->Client heap management. The framework will call your maker and killers as appropriate.

Device *InverterMaker() { return new Inverter; };
void InverterKiller(Device *spr) { delete spr; };

// What you MUST DO in the constructor of your device:
// 1) Set the Nametag, Maker and Killer in the lifeptrs struct.
// What you SHOULD do in the constructor:
// 2) Call SetLinks() with the # of inputs and outputs your device is going to use
//      ...if you dont do #2, you will use the default # of inputs your parent class assigns
Inverter::Inverter(void)
{
	lifeptrs.maker = InverterMaker;
	lifeptrs.killer = InverterKiller;
	strncpy(lifeptrs.nametag, "EXIV", 4);
	SetLinks(1, 1);
// This is how you would add a parameter to a device that is managed by the generic dlg system
// This adds a floating point parameter. You can add floats, ints, bools, or enums.
	AddParam(Parameter("Dummy Parameter", 0.0f, 0.0f, 1.0f));
}

// This is a convenient way to access your device parameters
// Note that accessing parameters is really quite slow due to the string searching for the param name. 
// Do NOT call inside of loops if you care about performance -- just retrieve the value once and cache it.
#define INV_DUMMY			ParmFRef("Dummy Parameter")


Inverter::~Inverter(void)
{

}

// You are REQUIRED to write out and read in your 4-byte Device ID tag, as shown below. It is also recommended that your device
// have an internal version number that you keep track of to allow backwards-compatibility if you ever change your Device save layout...
// Also, make sure you call your super class's Load and Save functions as shown here. Failure to do so will cause major problems.
bool Inverter::Load(std::istream &in) {
	char tag[5];
	in.read( tag, 4);
	if (strncmp(tag, lifeptrs.nametag, 4) == 0) {
		int ver = 0;
		in.read((char*) &ver, 1);
		return Filter::Load(in);
	}
	else
		return false;
};

bool Inverter::Save(std::ostream &out) {
	out.write( lifeptrs.nametag, 4);
	int ver = 1;
	out.write((char*) &ver, 1);
	return Filter::Save(out);
};

// The Activate function is where you work your magic. You should do four things in the Activate function:
// 1) Retrieve your inputs by calling RetrieveData(input#)
// 2) Do whatever you will do on it...
// 3) Store the data to make it available to your outputs
// 4) Return true to indicate successful activation, or false if there was an error.
bool Inverter::Activate(BuildContext &context) {
	HField *hf = HF(RetrieveData(0, context) );	// Use the RetrieveData(),StoreData() pair to get and set your heightfields

	// This is important! An input CAN RETURN NULL if it is not connected, or the input is turned off. So always check and bail out if its not there
	if (!hf)
		return false;

// ** example: if we actually DID anything with our dummy param, this is how you would get it
	float dummy = INV_DUMMY;

// ** example: If you wanted to access the data by coordinate instead of by index, here's how you would do it:
	int width = hf->w();
	int height = hf->h();
	float dummy_height = (*hf)[Coord(0,0)];	// use the coordinate version of the hf array operator

// heights are always in the range 0..1. So the inverter can just do height' = 1 - height
	for (int i = 0; i < hf->area(); i++) {
		(*hf)[i] = 1.0f - (*hf)[i];
	}

// Handoff to output(s)
	StoreData(hf,0, context); // pass the heightfield to our output stage.
// Success!
	return true;
}
